/*
**  GETIP - a simple "what's my IP address" program with addictinal PING capabilities
**  Copyright (C) 1999 Lapo Luchini <lapo@lapo.it>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
**  Revision 1.00 1999/05/22 20:56 GMT+1
**    first release
**  Revision 1.01 1999/05/23 07:45 GMT+1
**    fix: creashed when GPL.TXT not found, now is somehow searched in the same directory of the .exe
**    new: option to print sources to screen
**  Revision 1.02 1999/05/24 19:12 GMT+1
**    new: much more comments in the source
**    fix: added the trailing zero the the string used to print license or source (see print_file func)
**
**  note from the author: I created this program to learn some TCP/IP and I made it apply the GPL because
**    I think that ALL software should be free. If you like it or find it useful, please send an e-mail:
**    there's no real reason, but some feedback is always appreciated :-)
**    BTW: I was intended to compile on a standard Win32 platform, such as CygWin with option -bno_cygwin
**    (go to http://sourceware.cygnus.com/cygwin/ to download the free CygWin compiler for Win32)
**    you can always find the latest version on http://www.lapo.it
*/

#define VER "1.02"
//#define DEBUG
//#define ANSI

#include <stdio.h>
#include <stdlib.h>
#define Win32_Winsock
#include <windows.h>
#include <winsock.h>

// ********* ANSI-Colors defs **********
#ifdef ANSI
  #define ANSI_DEFAULT "\033[0m"
  #define A0 "\033[0;36m"
  #define A1 "\033[1;36m"
#else
  #define ANSI_DEFAULT ""
  #define A0 ""
  #define A1 ""
#endif

// ********* Version defs **********
#define TITLE "GETIP " VER " (c) 1999 Lapo Luchini <lapo@lapo.it>"
#define HELP A1 TITLE A0 "\n" \
             "This software comes with ABSOLUTELY NO WARRANTY. This is free software, and you\n" \
             "are welcome to redistribute it under certain conditions.\n" \
             "\n" \
             "GETIP.EXE -h (to obtain this help)\n" \
             "GETIP.EXE -l (to print GPL.TXT, containing the General Public License version 2)\n" \
             "GETIP.EXE -c (to print GETIP.C, containing the full sources)\n" \
             "GETIP.EXE [-n number_of_pings(0-65536)] [-s packet_size(12-65536)] [host]\n" \
             "known bugs: doesn't timeout waiting for ECHOREPLY"

// ********* IP&ICMP defs **********

// IP header used in ICMP packets
struct ip {
  BYTE hlen:4, // little endian, exgange order if big endian
       ver:4;
  WORD len;
  UINT id;
  WORD fragoff;
  BYTE ttl;
  BYTE proto;
  UINT chksum;
  IN_ADDR src_addr;
  IN_ADDR dst_addr;
};

// ICMP header
struct icmp { // note: sizeof(icmp) MUST be 8, as somewhere in the code is supposed to
  BYTE type;   // (may be generalized, but anyway the correct structure HAS 8 bytes so...)
  BYTE code;
  WORD cksum;
  WORD id;
  WORD seq;
};

// values for icmp.type
#define ECHOREPLY          0
#define ECHO               8

// Used to calculate the checksum of the headers. When calculating it the checksum field itself must be 0
WORD InternetChkSum(LPWORD data, WORD len) {
  long s=0;
  WORD o;
  while(len>1) {
    s+=*data++;
    len-=2;
  }
  if(len==1) {
    o=*(LPBYTE)data;
    s+=o;
  }
  s=(s>>16)+(s&0xFFFF);
  s+=(s>>16);
  return(~s);
}

// Just prints the file A, taking the path from B, to the screen
// ex. print_file("C:\Utility\GETIP.EXE", "GPL.TXT") prints C:\Utility\GPL.TXT
// basically it appends A to the end of (a copy of) B, starting after the last '\' or '/'
// I use it giving av[0] as the A parameters, but OSes that don't give full path in av[0] would need
// a much more complicated function (or the inclusion of some more libraries - this choise was biased
// toward smallness)
// A simple, and irritating, case in which this bad behaviour show itself is compiling with GNU C and
// using GETIP.EXE from the path: av[0] contains only the name of the executable and NOT the full path.
void print_file(char b[], char a[]) {
  FILE *in;
  char base[200];
  int c, l=0;
  for(c=0; b[c]!=0; c++) { // copies B into BASE and searchs the last '\' or '/' (stored in L)
    base[c]=b[c];
    if((base[c]=='\\')||(base[c]=='/'))
      l=c+1;
  }
  for(c=0; a[c]!=0; ) // copies A ofer the last part of B, overriding the old filename with the new one
    base[l++]=a[c++];
  base[l]=0;
  #ifdef DEBUG
    printf(A0 "Base:     " A1 "%s\n", b);
    printf(A0 "Filename: " A1 "%s\n", a);
    printf(A0 "New File: " A1 "%s\n", base);
  #endif
  if((in=fopen(base, "rt"))==NULL) {
    // just some pretty message insted of a simple "FILE NOT FOUND"
    printf("Couln't find the file GPL.TXT!\n" \
           "Refer to http://www.gnu.org/copyleft/gpl.txt to obtain a copy.");
  } else {
    while((c=fgetc(in))!=EOF) // copies the file into stdout
      printf("%c", c); // using putchar gives strange errors under -mno-cygwin mode, but this is SO slow
    fclose(in);
  }
}

int main(int ac, char *av[]) {
  LPHOSTENT l;
  WSADATA wsad;
  SOCKET hs;
  DWORD t;
  PIN_ADDR *al;

  // ********** parse parameters **********
  DWORD dotted=0;
  CHAR searchaddr[80]="";
  DWORD numpings=0;
  DWORD packet_oh=0;
  #ifdef DEBUG
    printf(A0 "ac=" A1 "%d\n", ac);
  #endif
  for(t=1; t<ac; t++) {
    #ifdef DEBUG
      printf(A0 "av[%d]=" A1 "%s\n", t, av[t]);
    #endif
    if(av[t][0]=='-') { //  un'opzione e c' il suo parametro
      switch(av[t][1]) {
        case 'h': printf(HELP ANSI_DEFAULT); // prints help (no more parameters needed)
                  return(0);
        case 'l': print_file(av[0], "gpl.txt"); // prints license (no more parameters needed)
                  return(0);
        case 'c': print_file(av[0], "getip.c"); // prints source (no more parameters needed)
                  return(0);
        case 'n': if(t<ac-1) { // number of pings to do (1 more parameter needed)
                    t++;
                    if((av[t][0]>='0')&&(av[t][0]<='9'))
                      sscanf(av[t], "%u", &numpings);
                    if(numpings>65536) {
                      printf(A0 "Invalid number of ping, using default of 0.\n");
                      numpings=1;
                    }
                    #ifdef DEBUG
                      printf(A0 "Paramentro numping=" A1 "%d\n", numpings);
                    #endif
                  }
                  break;
        case 's': if(t<ac-1) { // packet overhead (1 more parameter needed)
                    t++;
                    if((av[t][0]>='0')&&(av[t][0]<='9'))
                      sscanf(av[t], "%u", &packet_oh);
                    if((packet_oh<12)||(packet_oh>65536)) {
                      printf(A0 "Invalid packet size, using minimum of 12.\n");
                      packet_oh=12;
                    }
                    packet_oh-=12;
                    #ifdef DEBUG
                      printf(A0 "Paramentro packet_oh=" A1 "%d\n", packet_oh);
                    #endif
                    packet_oh>>=2; // number of DWORD (starting from BYTEs)
                  }
                  break;
      }
    } else if((av[t][0]>='0')&&(av[t][0]<='9')) { //  un indirizzo dotted
      dotted=inet_addr(av[t]);
      #ifdef DEBUG
        printf(A0 "Indirizzo dotted=" A1 "%s\n", inet_ntoa(*(LPIN_ADDR)(&dotted)));
      #endif
    } else {
      int c;
      for(c=0; (av[t]!=0)&&(c<80); c++)
        searchaddr[c]=av[t][c];
      searchaddr[79]=0; // closes the string in case it was (strangely) longer
      #ifdef DEBUG
        printf(A0 "Indirizzo letterale=" A1 "%s\n", searchaddr);
      #endif
    }
  }
  #ifdef DEBUG
    printf(A0 "fine parse deiparametri.\n");
    printf(A0 "Paramentro numping=" A1 "%d\n", numpings);
    printf(A0 "Paramentro packet_oh=" A1 "%d\n", packet_oh);
  #endif

  printf(ANSI_DEFAULT);

  SetConsoleTitle(TITLE); // set the window's title (if not fullscreen)

  // ********** begin of main part **********
  if(WSAStartup(0x0101, &wsad)) {
    printf("Could not inizialize WSOCK32.DLL\n");
    exit(1);
  }
  if(dotted) // if address was given in dotted numeric form
    l=gethostbyaddr((char *)&dotted, 4, AF_INET); // ..use gethostbyaddr
  else if(searchaddr[0]) // else it bust have been in string form
    l=gethostbyname(searchaddr); // ..then use gethostbyname
  else // is first char was a 0 (empty string)
    l=gethostbyname(NULL); // ..then use same, but with no parameters (gives current machine infos)
  if (!l)
    printf(A0 "Could not get " A1 "IP address" A0 ".\n");
  else {
    printf(A0 "HOST Name: " A1 "%s\n", l->h_name);
    t=0;
    while(l->h_aliases[t])
      printf(A0 "    Alias: " A1 "%s\n", l->h_aliases[t++]);
    al=(PIN_ADDR *)l->h_addr_list;
    while(*al) {
      union {
        struct {
          unsigned a:8,
                   b:8,
                   c:8,
                   d:8;
        } b;
        unsigned long l;
      } ad;
      char c='?';
      long m=0x00000000;
      ad.l=ntohl((**al).s_addr);
      if(IN_CLASSA(ad.l)) {
        c='A';
        m=IN_CLASSA_NET;
      } else if(IN_CLASSB(ad.l)) {
        c='B';
        m=IN_CLASSB_NET;
      } else if(IN_CLASSC(ad.l)) {
        c='C';
        m=IN_CLASSC_NET;
      }
      printf(A0 "IP:" A1 "%03d.%03d.%03d.%03d" A0 " Hex:" A1 "%08lX" A0 \
             " Class:" A1 "%c" A0 " NetMask:" A1 "%lX\n",
             ad.b.d, ad.b.c, ad.b.b, ad.b.a,
             ad.l, c, m);
      al++;
    }
  }
  if(numpings&&(l||dotted)) { // if a host was found OR it was given in dotted form, then ping it!
    struct icmp *pic;
    BYTE pics[65536],
         picr[65536];
    int sb, rb, size;
    SOCKADDR_IN sal, sah;
    DWORD trip;
    DWORD c;
    hs=socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
    if(hs==INVALID_SOCKET) {
      printf(A0 "Couldn't open a RAW socket." ANSI_DEFAULT);
      return(1);
    }
    sal.sin_family=AF_INET;
    if(dotted)
      sal.sin_addr=*(LPIN_ADDR)(&dotted);
    else
      sal.sin_addr=*((LPIN_ADDR)*l->h_addr_list); // indirizzo appena cercato, da pingare
    pic=(struct icmp *)pics;
    pic->type=ECHO;
    pic->code=0;
    pic->id=0x0777;
    pic->seq=0;
    while (numpings--) {
      pic->cksum=0;
      *(PDWORD)(pics+sizeof(struct icmp))=GetTickCount(); // stores clock in the optional data field
      size=sizeof(struct icmp)+sizeof(DWORD); // MUST be 12, but don't know how to check at compile time
      t=0xFFFFFFFF;
      for(c=1; c<=packet_oh; c++) { // stores also some more 'random' data in there (as reqeusted)
        *(PDWORD)(pics+sizeof(struct icmp)+4*c)=t;
        t-=0x01357BDF;
      }
      size+=packet_oh*4;
      pic->cksum=InternetChkSum((LPWORD)pic, size);
      if(!pic->cksum) { // checksum MUST be != 0, as it is it's default value
        printf(A0 "CheckSum=0!" ANSI_DEFAULT);
        return(1);
      }
      sb=sendto(hs, (LPSTR)pics, size, 0, (LPSOCKADDR)&sal, sizeof(sal));
      printf(A0 "Ping... "); // if ctrl-break is pressed waiting for pong, then color doesn't reset
      if(sb==SOCKET_ERROR)
        printf(A0 "Socket Error Sending!\n");
      else if (sb!=size)
        printf(A0 "Wrong SentBytes!\n");
      else {
        int alen=sizeof(sah), iphl;
        struct ip *pip;
        #ifdef DEBUG
          printf(A0 "Inviati %d bytes.\n", sb);
          printf(A0);
          for(c=1; c<=sb; c++)
            printf("%02X%s", pics[c-1], (c%4 ? " " : (c%16 ? "  " : "\n")));
          printf(A0 "\nAspetto... ");
        #endif
        // next ione is a BLOCKING CALL! don't know how to easily modify it to have some sort of timeout
        rb=recvfrom(hs, (LPSTR)picr, sizeof(picr), 0, (LPSOCKADDR)&sah, &alen);
        trip=GetTickCount(); // gets new time
        pip=(struct ip *)picr;
        iphl=pip->hlen<<2;
        #ifdef DEBUG
          printf("HLEN:%X VER:%X\n", pip->hlen, pip->ver);
        #endif
        if(rb==SOCKET_ERROR)
          printf(A0 "Socket Error Receiving!\n");
        else {
          if (rb<(iphl+sizeof(struct icmp)))
            printf("[Too Short] ");
          #ifdef DEBUG
            printf("Ricevuto (%d/%d bytes)!\n", rb, iphl+sizeof(struct icmp));
          #endif
          pic=(struct icmp *)(picr+iphl);
          if(pic->type!=ECHOREPLY)
            printf("Was not an ECHOREPLY [%d]\n", pic->type);
          else if(pic->id!=0x0777)
            printf("Was not mine\n");
          else {
            WORD errs=0;
            for(c=sizeof(struct icmp); (c<sb)&&(c+iphl<rb); c++)
              if(pics[c]!=picr[c+iphl])
                errs++;
            printf(A0 "pong from " A1 "%s" A0 " in" A1 "%4u" A0 " ms (" A1 "%d" A0 "sent/" A1 "%d" A0 "rcvd/" A1 "%d" A0 "errs bytes).\n",
                   inet_ntoa(sah.sin_addr), trip-*(PDWORD)(pics+sizeof(struct icmp)),
                   sb, rb-iphl, errs);
          }
          #ifdef DEBUG
            printf(A0);
            for(c=1; c<=rb; c++)
              printf("%02X%s", picr[c-1], (c%4 ? " " : (c%16 ? "  " : "\n")));
            printf(A1 "\n");
          #endif
        }
      }
    }
    closesocket(hs);
  }
  WSACleanup();

  printf(ANSI_DEFAULT);

  return(0);
}
